/* written by John Doug Reynolds, February 2002 */

#include "rgsfluxer.h"
#include <rgsPixelsLib.h>
#include <DalVec.h>
#include <cstdio>
#include <cmath>

using namespace rgsPixelsLib;

const float hc = 12.398; // [keV Angstrom] energy times wavelength

class BetaToLambda {
public:
   BetaToLambda( float alphaNom, float lineSep, unsigned order );
   float operator()( float beta ) const { return _slope * cos(beta) + _icept; }
private:
   float _slope, _icept;
};

class ResponseMatrix {
public:
   ~ResponseMatrix();
   ResponseMatrix( const Table* );
   pair<float,float> changeRow( unsigned );
   bool next( short &channel, float &value );
   unsigned rows() const { return _rows; }
private:
   const DalVecI _nGrp;
   const DalVecE _eLow, _eHigh;
   const Column *_fChan, *_nChan, *_matrix;
   const CellData *_fChanData, *_nChanData, *_matrixData;
   const short *_fChanPtr, *_nChanPtr;
   const float *_matrixPtr;
   unsigned _groupCount, _itemCount, _rows;
   short _channel;
};

class Fluxer {
public:
   enum AxisMode { Wavelength, Energy };

   Fluxer( enum AxisMode, const GridSpan&, const vector<int> &qualities );

   void add( const string &pha, const string &rmf );

   bool at( int cell, float &chan, float &flux, float &err, int &qual ) const;

private:
   enum AxisMode _mode;
   struct GridSpan _grid;
   vector<float> _rate, _err2, _area;
   vector<int> _quality;
   unsigned _qualityRank[5];

   static void rebin( vector<float> &vec, float min, float max, float val );

   float binWidth( const pair<float,float>& ) const;
   float energyToGrid( float energy ) const;
   float lambdaToGrid( float lambda ) const;
   unsigned addSpectrum( const string &filename, vector<bool> &bad );
   void addResponse( const string &filename,
		     unsigned spectrumOrder, const vector<bool> &bad );
};

void rgsfluxer()
{
   string filename = stringParameter("file");
   string format = stringParameter("format");
   bool withaltnan = booleanParameter("withaltnan");
   float nan = realParameter("nan");

   vector<string> pha = stringListParameter("pha");
   vector<string> rmf = stringListParameter("rmf");
   vector<int> qualities = intListParameter("quality");

   struct GridSpan grid( realParameter("min"),
			 realParameter("max"),
			 intParameter("bins")
			 );

   enum Fluxer::AxisMode mode=(Fluxer::AxisMode)-1;
   string modeStr = stringParameter("mode");
   const char *binUnits=0, *valUnits=0;
   if ( modeStr == "wavelength" ) {
      mode = Fluxer::Wavelength;
      binUnits = "Angstrom";
      valUnits = "1/(s cm^2 A)";
   }
   else if ( modeStr == "energy" ) {
      mode = Fluxer::Energy;
      binUnits = "keV";
      valUnits = "1/(s cm^2 keV)";
   }
   else Sas::error("badMode","illegal value specified for mode parameter");

   if ( pha.size() != rmf.size() )
      Sas::error("pairing","unpaired pha or rmf");

   //

   struct Fluxer fluxer( mode, grid, qualities );

   vector<string>::const_iterator phaI = pha.begin(), rmfI = rmf.begin();

   for ( string defaultRmf; rmfI != rmf.end(); ++rmfI ) {
      string rmfName;

      if ( *rmfI == "@" ) {
	 rmfName = *phaI;
	 string::size_type size = rmfName.size();
	 if ( size < 14 || rmfName.substr(size-14,6) != "SRSPEC" )
	    Sas::error("badPhaName","cannot infer corresponding rmf");
	 rmfName.replace( size-14, 6, "RSPMAT" );
      }
      else if ( *rmfI == "." ) {
	 if ( defaultRmf == "" )
	    Sas::error("illegalRmfName","first rmf cannot be '.'");
	 rmfName = defaultRmf;
      }
      else
	 rmfName = *rmfI;

      fluxer.add( *phaI++, rmfName );
      defaultRmf = rmfName;
   }

   //

   bool withquality = (qualities.size() > 1);

   if ( format == "dal" ) {
      if ( filename == "" ) filename = "flux.ds";
      struct DataSet *dataset = dataSetServer->open( filename, Dal::Create );
      struct Table *table = dataset->addTable("FLUXED",grid.length());
      {
	 DalVecE channel( table->addColumn("CHANNEL",Dal::Real32,
					   modeStr.c_str(),binUnits) );
	 DalVecE flux( table->addColumn("FLUX",Dal::Real32,
					"observed flux",valUnits) );
	 DalVecE error( table->addColumn("ERROR",Dal::Real32,
					 "statistical uncertainty",valUnits) );
	 DalVecI quality;
	 if ( withquality )
	    quality.reset( table->addColumn("QUALITY",Dal::Int16,
					    "quality flag") );

	 for ( int c = grid.min; c <= grid.max; ++c ) {
	    int qual, i = c - grid.min;
	    if ( !fluxer.at(c,channel[i],flux[i],error[i],qual) )
	       if ( !withaltnan ) {
		  flux.setToNull(i);
		  error.setToNull(i);
	       }
	       else flux[i] = error[i] = nan;
	    if ( withquality ) quality[i] = qual;
	 }
      }
      dataSetServer->close( dataset );
   }
   else {
      if ( filename == "" )
	 filename = format == "qdp" ? "flux.qdp" : "flux.txt";
      const char *snan = format == "qdp" ? "no" : "NaN";
      FILE *out = fopen( filename.c_str(), "w" );
      if ( format == "qdp" ) fprintf( out, "READ SERR 2\n" );
      fprintf( out, "!%14s%15s%15s%10s\n", modeStr.c_str(),
	       "flux", "error", withquality ? "quality" : "" );
      fprintf( out, "!%14s%15s%15s\n", binUnits, valUnits, valUnits );
      for ( int c = grid.min; c <= grid.max; ++c ) {
	 float chan, flux = nan, err = nan;
	 int qual;
	 if ( fluxer.at(c,chan,flux,err,qual) || withaltnan )
	    fprintf( out, "%15e%15e%15e", chan, flux, err );
	 else
	    fprintf( out, "%15e%15s%15s", chan, snan, snan );
	 if ( withquality )
	    fprintf( out, "%10d", qual );
	 fprintf( out, "\n" );
      }
      fclose( out );
   }
}

/* --- BetaToLambda methods ------------------------------------------------ */

BetaToLambda::BetaToLambda( float alphaNom, float lineSep, unsigned order )
{
   _slope = -lineSep / order;
   _icept = -cos(alphaNom) * _slope;
}

/* --- ResponseMatrix methods ---------------------------------------------- */

ResponseMatrix::~ResponseMatrix()
{
   delete _matrixData;
   delete _nChanData;
   delete _fChanData;
}

ResponseMatrix::ResponseMatrix( const Table *table )
   : _nGrp( table->column("N_GRP") ),
     _eLow( table->column("ENERG_LO") ),
     _eHigh( table->column("ENERG_HI") ),
     _fChan( table->column("F_CHAN") ),
     _nChan( table->column("N_CHAN") ),
     _matrix( table->column("MATRIX") ),
     _fChanData(0), _nChanData(0), _matrixData(0),
     _groupCount(0), _itemCount(0),
     _rows( table->rows() )
{}

pair<float,float> ResponseMatrix::changeRow( unsigned row )
{
   delete _matrixData;
   delete _nChanData;
   delete _fChanData;

   _fChanData = _fChan->cellData(row);
   _nChanData = _nChan->cellData(row);
   _matrixData = _matrix->cellData(row);

   _fChanPtr = _fChanData->int16Data();
   _nChanPtr = _nChanData->int16Data();
   _matrixPtr = _matrixData->real32Data();

   _groupCount = _nGrp[row];
   _itemCount = 0;

   return pair<float,float>(_eLow[row],_eHigh[row]);
}

bool ResponseMatrix::next( short &channel, float &value )
{
   for (;;)
      if ( _itemCount ) {
	 channel = _channel++;
	 value = *_matrixPtr++;
	 -- _itemCount;
	 return true;
      }
      else if ( _groupCount ) {
	 _channel = *_fChanPtr++ - 1;
	 _itemCount = *_nChanPtr++;
	 -- _groupCount;
      }
      else return false;
}

/* --- Fluxer methods ------------------------------------------------------ */

void Fluxer::rebin( vector<float> &vec, float min, float max, float val )
{
   if ( max < min ) swap(min,max);

   if ( min >= vec.size() || max <= 0 ) return;

   if ( min < 0 ) {
      val *= 1 + min/(max - min);
      min = 0;
   }

   if ( max > vec.size() ) {
      val *= 1 - (max - vec.size())/(max - min);
      max = vec.size();
   }

   if ( floor(max) > floor(min) ) {
      float portion = val / (max - min);
      if ( ceil(min) > min ) {
	 vec[int(floor(min))] += portion * (ceil(min) - min);
	 min = ceil(min);
      }
      while ( floor(max) > floor(min) ) {
	 vec[int(floor(min))] += portion;
	 min += 1;
      }
      vec[int(floor(min))] += portion * (max - min);
   }
   else
      vec[int(floor(min))] += val;
}

Fluxer::Fluxer( enum AxisMode mode, const GridSpan &grid,
		const vector<int> &qualities )
   : _mode(mode), _grid(grid),
     _rate(grid.length()), _err2(grid.length()),
     _area(grid.length()), _quality(grid.length())
{
   unsigned i;
   vector<int>::const_iterator iter;
   for ( i = 0; i < 5; ++i ) _qualityRank[i] = 0;
   for ( i = 0, iter = qualities.begin(); iter != qualities.end(); ++iter )
      _qualityRank[*iter] = ++i;
}

void Fluxer::add( const string &pha, const string &rmf )
{
   vector<bool> bad;
   addResponse( rmf, addSpectrum(pha,bad), bad );
}

bool Fluxer::at( int c, float &chan, float &flux, float &err, int &qual ) const
{
   int i = c - _grid.min;
   float area = _area[i];
   if ( area ) {
      flux = _rate[i] / area;
      err = sqrt( _err2[i] ) / area;
   }
   qual = _quality[i];
   chan = _grid[c];
   return bool(area);
}

float Fluxer::binWidth( const pair<float,float> &limits ) const
{
   return _mode == Energy
      ? limits.second - limits.first : hc/limits.first - hc/limits.second;
}

float Fluxer::energyToGrid( float energy ) const
{
   return _grid.tick( _mode == Energy ? energy : hc / energy );
}

float Fluxer::lambdaToGrid( float lambda ) const
{
   return _grid.tick( _mode == Wavelength ? lambda : hc / lambda );
}

inline float sqr( float x ) { return x * x; }
unsigned Fluxer::addSpectrum( const string &filename, vector<bool> &bad )
{
   struct DataSet *dataset = dataSetServer->open(filename,Dal::Read);
   const Table *table = dataset->table("SPECTRUM");
   unsigned order = table->attribute("RFLORDER")->asInt();
   float exposure = table->attribute("EXPOSURE")->asReal();
   bool gross = table->attribute("HDUCLAS2")->asString() != "NET";

   bad.resize( table->rows(), false );

   struct BetaToLambda toLambda( table->attribute("RGSALPHA")->asReal(),
				 table->attribute("LINE_SEP")->asReal(),
				 order );

   {
      const DalVecI channel( table->column("CHANNEL") );
      const DalVecJ counts( table->column("COUNTS") );
      const DalVecE areascal( table->column("AREASCAL") );
      const DalVecI quality( table->column("QUALITY") );
      DalVecE stat_err;
      if ( !gross ) stat_err.reset( table->column("STAT_ERR") );

      struct GridSpan beta = gridspan( channel.column() );

      for ( unsigned i = 0; i < table->rows(); ++i )
	 if ( _qualityRank[ quality[i] ] ) {
	    float min = lambdaToGrid( toLambda(beta[channel[i]-.5]) ) + .5;
	    float max = lambdaToGrid( toLambda(beta[channel[i]+.5]) ) + .5;

	    float time = exposure * areascal[i];
	    float rate = counts[i] / time;
	    float err2 = gross ? rate / time : sqr(stat_err[i] / time);

	    rebin( _rate, min, max, rate );
	    rebin( _err2, min, max, err2 );

	    struct Span span = Span(int(floor(min)),int(floor(max))) & _grid;
	    for ( int c = span.min; c <= span.max; ++c ) {
	       int j = c - _grid.min;
	       if ( _qualityRank[ _quality[j] ] < _qualityRank[ quality[i] ] )
		  _quality[j] = quality[i];
	    }
	 }
	 else bad[i] = true;
   }

   dataSetServer->close( dataset );
   return order;
}

void Fluxer::addResponse( const string &filename,
			  unsigned spectrumOrder, const vector<bool> &bad )
{
   struct DataSet *dataset = dataSetServer->open(filename,Dal::Read);
   const Table *btab = dataset->table("EBOUNDS");

   if ( int(spectrumOrder) != dataset->attribute("RFLORDER")->asInt() )
      Sas::error("wrongOrder",
		 "response matrix order not same as spectrum");

   {
      const DalVecI channel( btab->column("CHANNEL") );
      for ( unsigned i = 1; i <= btab->rows(); ++i )
	 if ( channel(i) != short(i) )
	    Sas::error("channelGaps",
		       "cannot handle response matrix with channel gaps");
   }

   {
      struct ResponseMatrix matrix( dataset->table("MATRIX") );
      const DalVecE eMin( btab->column("E_MIN") );
      const DalVecE eMax( btab->column("E_MAX") );
      for ( unsigned i = 0; i < matrix.rows(); ++i ) {
	 short channel;
	 float value, delta = binWidth( matrix.changeRow(i) );
	 while ( matrix.next(channel,value) )
	    if ( ! bad[channel] )
	       rebin( _area,
		      energyToGrid( eMin[channel] ) + .5,
		      energyToGrid( eMax[channel] ) + .5,
		      delta * value
		      );
      }
   }

   dataSetServer->close( dataset );
}
